//////////////////////////////////////////////////////////////////////////////// 
/// Pseudoglobe Screensaver
/// 
/// Copyright  2002  Sebastian Pipping <webmaster@hartwork.org>
/// 
/// -->  http://www.hartwork.org
/// 
/// This source code is released under the GNU General Public License (GPL).
/// See GPL.txt for details. Any non-GPL usage is strictly forbidden.
////////////////////////////////////////////////////////////////////////////////


#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <pbt.h>

#include <gl/gl.h>
#include <gl/glu.h>
#include <math.h>				// sin() and cos()
#include <shellapi.h>			// ShellAbout

#include "capplog.h"			// Application logger
#include "cregistry.h"			// Registry
#include "csaver.h"
#include "ctimer.h"

#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL 0x020A
#endif

#define APP_TITLE "Pseudoglobe Screensaver"
#define WNDCLASS_NAME "GLass"

#define PI 3.14159f

// Additional keys
#define VK_P 80
#define VK_T 84


// Speed adjustment
const float ANIMATION_STEP_DEFAULT = 0.7f;
const float ROTATION_STEP_DEFAULT = 0.00591f;
float fAnimationStep = ANIMATION_STEP_DEFAULT;
float fRotationStep = ROTATION_STEP_DEFAULT;

// Winamp
#define WM_WA_IPC WM_USER
#define IPC_GETVERSION 0

HWND hWinamp;
BOOL bWinampFound = FALSE;
int iWinampVersionMajor;
int iWinampVersionMinor;

// Settings
BOOL bBlending				= FALSE;		// Alpha blending
const BOOL bAlternateColor	= TRUE;			// Alternate color mode
const BOOL bGrayscale		= FALSE;		// Grayscale mode
const int MIN_MOUSE_MOVE	= 50;

// Internal
BOOL bKeys[ 256 ];
POINT MousePosFirst;
BOOL bActive = TRUE;

// Powersave
BOOL bPowersaveMonitor = FALSE;
BOOL bPowersaveSuspension = FALSE;

CAppLog MyLog;
CRegistry MyReg;
CSaver MySaver;
CTimer MyTimer;

HDC			hDC = NULL;			// Device Context
HGLRC		hRC = NULL;			// Rendering Context
HWND		hWnd = NULL;		// Window Hanlde
HINSTANCE	hGlobalInstance;	// Application Instance

// Essential prototypes
LRESULT	CALLBACK WndProc( HWND, UINT, WPARAM, LPARAM );

// Degree to radiant
double deg2rad( double deg )
{
	return deg / 180 * PI;
}

int abs( int x )
{
	return ( x > 0 ) ? x : -x;	
}

// Color a polygon
void ColorPart( int iAll, int iBrightness, int iColor )
{
	float fRed, fGreen, fBlue;

	if( !bAlternateColor )
	{
		float fSection = ( float )iColor / iAll / 2 * 3;

		// Red
		fRed = 0.0f;
		if( fSection < 1 )
		{
			fRed = 1.0f - fSection;
		}
		else if( fSection > 2 )
		{
			fRed = fSection - 2.0f;
		}

		// Green
		fGreen = 0.0f;
		if( fSection <= 1 )
		{
			fGreen = fSection;
		}
		else if( fSection < 2 ) // && ( fSection > 1 )
		{
			fGreen = 2.0f - fSection;
		}

		// Blue
		fBlue = 0.0f;
		if( fSection > 2 )
		{
			fBlue = 3.0f - fSection;
		}
		else if( fSection > 1 ) // && ( fSection <= 2 )
		{
			fBlue = fSection - 1.0f;
		}
	}
	else
	{
		float fSection = ( float )iColor / iAll / 2;
		fRed	= ( float )asin( sin( fSection * PI ) ) * 2 / PI;
		fGreen	= ( float )fabs( asin( sin( fSection * PI + PI / 3 ) ) * 2 / PI );
		fBlue	= ( float )fabs( asin( sin( fSection * PI + 2 * PI / 3) ) * 2 / PI );
	}

	if( bGrayscale )
	{
		float fGray = ( fRed + fGreen + fBlue ) / 3.0f;
		fRed = fGreen = fBlue = fGray;
	}
	
	const int iMiddle = ( iAll - 1 ) / 2;
	const float fFactor = ( float )( iBrightness - iMiddle ) / iMiddle ;
	if( iBrightness > iMiddle )
	{
		// Brighten
		fRed += ( 1.0f - fRed ) * fFactor;
		fGreen += ( 1.0f - fGreen ) * fFactor;
		fBlue += ( 1.0f - fBlue ) * fFactor;
	}
	else
	{
		// Darken
		fRed += fRed * fFactor;
		fGreen += fGreen * fFactor;
		fBlue += fBlue * fFactor;
	}

	glColor3f( fRed, fGreen, fBlue );
}

int GetIndexX( int iIndex, int iAll )
{
	if( iIndex % 2 )
	{
		// Odd number
		return iAll - ( iIndex + 1 ) / 2;
	}
	else
	{
		// Even number
		return iAll + iIndex / 2;
	}
}

int GetIndexY( int iIndex, int iAll )
{
	if( iIndex % 2 )
	{
		// Odd number
		return iAll - ( iIndex + 1 ) / 2;
	}
	else
	{
		// Even number
		return iIndex / 2;
	}
}

// Create main object
void CreateObjectList( int iParts )
{
	glNewList( 10, GL_COMPILE );
	
		glBegin( GL_QUADS );

			// GLOBE
			for( int j = 0; j < iParts * 2; j++ )
			{
				const int j2 = GetIndexX( j, iParts );

				const double dAng2 = deg2rad( j2 * 180 / iParts );
				const double dAng22 = deg2rad( ( j2 + 1 ) * 180 / iParts );
				for( int i = 0; i < iParts; i++ )
				{
					const int i2 = GetIndexY( i, iParts );

					ColorPart( iParts, i2, j2 );
				
					const double dAngle1 = deg2rad( i2 * 180 / iParts );
					const double dAngle12 = deg2rad( ( i2 + 1 ) * 180 / iParts );
					
					glVertex3f( ( float )sin( dAng2 ) * ( float )sin( dAngle1 ), ( float )cos( dAngle1 + PI ), ( float )cos( dAng2 ) * ( float )sin( dAngle1 ) );
					glVertex3f( ( float )sin( dAng2 ) * ( float )sin( dAngle12 ), ( float )cos( dAngle12 + PI ), ( float )cos( dAng2 ) * ( float )sin( dAngle12 ) );
					glVertex3f( ( float )sin( dAng22 ) * ( float )sin( dAngle12 ), ( float )cos( dAngle12 + PI ), ( float )cos( dAng22 ) * ( float )sin( dAngle12 ) );
					glVertex3f( ( float )sin( dAng22 ) * ( float )sin( dAngle1 ), ( float )cos( dAngle1 + PI ), ( float )cos( dAng22 ) * ( float )sin( dAngle1 ) );
				}
			}

		glEnd();

	glEndList();
	MyLog.Append( "Globe object created (%i polygons)\n", iParts * iParts * 2 );
}

void ChooseDepthTestFunction( BOOL bUseBlending )
{
	if( bUseBlending )
	{
		glDepthFunc( GL_ALWAYS );
	}
	else
	{
		glDepthFunc( GL_LEQUAL );
	}
}

void EnableBlending( BOOL bUseBlending )
{
	if( bUseBlending )
	{
		glEnable( GL_BLEND );
		MyLog.Append( "ALPHA BLENDING enabled\n" );
	}
	else
	{
		glDisable( GL_BLEND );
		MyLog.Append( "ALPHA BLENDING disabled\n" );
	}
}

int Initialize()
{
	// Enable smooth shading
	glShadeModel( GL_SMOOTH );
	//glShadeModel( GL_FLAT );

	// Set blackground color
	glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );

	// Set depth buffer
	glClearDepth( 1.0f );

	// Enable depth testing
	glEnable( GL_DEPTH_TEST );
	ChooseDepthTestFunction( bBlending );

	// Enable blending
	EnableBlending( bBlending );
	glBlendFunc( GL_SRC_ALPHA, GL_DST_ALPHA );

	//glEnable( GL_POLYGON_SMOOTH );
	//glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );


	//glDisable(GL_DEPTH_TEST);

	// Select best perspektive corrections
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

	// Create object
	CreateObjectList( 15 );

	return TRUE;
}

void ResizeScene( GLsizei iWidth, GLsizei iHeight )
{
	// Prevent divide by zero
	if( !iHeight ) iHeight++;

	// Reset viewport
	glViewport( 0, 0, iWidth, iHeight );

	// Reset projection matrix
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	// Set perspective view
	gluPerspective( 45.0f,( GLfloat )iWidth / ( GLfloat )iHeight, 0.1f, 100.0f );

	// Reset modelview matrix
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	MyLog.Append( "Scene resized to %ix%i\n", iWidth, iHeight );
}

void DrawScene()
{
	static float alpha = 0.0f;
	static float t = 0.0f;

	// Clear screen and depth buffer
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// Reset modelview matrix
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	// Move into scene
	//glTranslatef( 1.0f + cos( alpha / 70 ), 0.0f, -4.0f + sin( alpha / 50 ) );
	glTranslatef( 0.0f, 0.0f, -4.0f );

	// Rotate object
	glRotatef( alpha, ( float )sin( t ), ( float )sin( t + 0.0022f * ROTATION_STEP_DEFAULT / fRotationStep ), ( float )cos( t + 0.00261f * ROTATION_STEP_DEFAULT / fRotationStep ) );
	t += fRotationStep;

	// Draw Object to buffer
	glCallList( 10 );

	// Animate
	alpha += fAnimationStep;
	if( alpha > 360 )
	{
		alpha -= 360;
	}

	// Print to window
	glFlush();
	SwapBuffers( hDC );
}

void SaverEngine()
{
	if( MyTimer.iFrames == 0 )
	{
		MyTimer.Start();
	}
	else if( MyTimer.AdjustFrame() )
	{
		MyTimer.Adjust();
	}
	MyTimer.iFrames++;

	DrawScene();
}

void KillWindow()
{
	// Return to previous mode
	if( MySaver.iMode == MODE_FULLSCREEN )
	{
		ChangeDisplaySettings( NULL,0 );
		MyLog.Append( "Original resolution restored\n" );
	}

	// Kill rendering context
	if( hRC )
	{
		// Release rendering context
		wglMakeCurrent( NULL, NULL );
		MyLog.Append( "Rendering context deactivated\n" );
		
		// Delete rendering context
		wglDeleteContext( hRC );
		hRC = NULL;
		MyLog.Append( "Rendering context deleted\n" );
	}

	// Kill device context
	if( hDC )
	{
		ReleaseDC( hWnd, hDC );
		hDC = NULL;
		MyLog.Append( "Device context released\n" );
	}

	// Kill window
	if( hWnd )
	{
		DestroyWindow( hWnd );
		hWnd = NULL;
		MyLog.Append( "Window destroyed\n" );
	}

	// Unregister class
	UnregisterClass( WNDCLASS_NAME, hGlobalInstance );
	MyLog.Append( "Window class unregistered\n" );

	// Delete instance
	hGlobalInstance = NULL;
}

BOOL OpenWindow()
{
	GLuint		PixelFormat;		// Pixel format
	WNDCLASS	wc;					// Window class structure
	DEVMODE		dm;					// Device mode

	int iWidth, iHeight, iBits;		// Window dimensions

	// Get current resolution
	dm.dmSize = sizeof( DEVMODE );
	EnumDisplaySettings( NULL, ENUM_CURRENT_SETTINGS, &dm );
	MyLog.Append( "Current screen resolution: %ix%i %i-Bit\n", dm.dmPelsWidth, dm.dmPelsHeight, dm.dmBitsPerPel );

	// Get window dimensions
	if( MySaver.iMode == MODE_FULLSCREEN )
	{
		iWidth = dm.dmPelsWidth;
		iHeight = dm.dmPelsHeight;
	}
	else
	{
		RECT ParentRect;
		GetWindowRect( MySaver.hParent, &ParentRect );
		iWidth = ParentRect.right - ParentRect.left;
		iHeight = ParentRect.bottom - ParentRect.top;
	}
	iBits = dm.dmBitsPerPel;

	// Describe window class
	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc		= ( WNDPROC )WndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= hGlobalInstance;
	wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO);					// Use window flag icon
	wc.hCursor			= LoadCursor( NULL, IDC_ARROW );				// Use arrow cursor
	wc.hbrBackground	= ( HBRUSH )GetStockObject( LTGRAY_BRUSH );		// Use Gray background
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= WNDCLASS_NAME;								// Name class

	// Register class
	if( !RegisterClass( &wc ) )
	{
		MyLog.Append( "Window class could not be registered\n" );
		return FALSE;
	}
	MyLog.Append( "Window class registered\n" );

/*
	// Switch to fullscreen mode
	if( MySaver.iMode == MODE_FULLCREEN )
	{
		DEVMODE dmScreenSettings;		// Device mode

		// Set device mode
		dmScreenSettings.dmSize			= sizeof( dmScreenSettings );
		dmScreenSettings.dmPelsWidth	= iWidth;
		dmScreenSettings.dmPelsHeight	= iHeight;
		dmScreenSettings.dmBitsPerPel	= iBits;
		dmScreenSettings.dmFields		= DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;

		// Change resolution
		if( ChangeDisplaySettings( &dmScreenSettings, CDS_FULLSCREEN ) != DISP_CHANGE_SUCCESSFUL )
		{
			bFullscreen = FALSE;
		}
	}
*/

	// Create window
	if( MySaver.iMode == MODE_FULLSCREEN )
	{
		hWnd = CreateWindowEx( WS_EX_APPWINDOW, WNDCLASS_NAME, APP_TITLE, WS_POPUP, 0, 0, iWidth, iHeight, NULL, NULL, hGlobalInstance, NULL );
	}
	else
	{
		hWnd = CreateWindow( WNDCLASS_NAME, APP_TITLE, WS_CHILD | WS_VISIBLE, 0, 0, iWidth, iHeight, MySaver.hParent, NULL, hGlobalInstance, NULL );
	}

	if( !hWnd )
	{
		MyLog.Append( "Window could not be created\n" );
		KillWindow();
		return FALSE;
	}
	MyLog.Append( "Window created\n" );

	// Configure pixel format
	static PIXELFORMATDESCRIPTOR pfd;
	pfd.nSize			= sizeof( PIXELFORMATDESCRIPTOR );
	pfd.nVersion		= 1;
	pfd.dwFlags			= PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType		= PFD_TYPE_RGBA;
	pfd.cColorBits		= iBits;
	pfd.cDepthBits		= 16;
	pfd.iLayerType		= PFD_MAIN_PLANE;
	
	// Get device context
	if( !( hDC = GetDC( hWnd ) ) )
	{
		MyLog.Append( "Could not get a device context\n" );
		KillWindow();
		return FALSE;
	}
	MyLog.Append( "Got a device context\n" );

	// Find matching pixel format
	if( !( PixelFormat = ChoosePixelFormat( hDC, &pfd ) ) )
	{
		MyLog.Append( "No matching pixel format found\n" );
		KillWindow();
		return FALSE;
	}
	MyLog.Append( "Matching pixel format found\n" );

	// Set pixel format
	if( !SetPixelFormat( hDC, PixelFormat, &pfd ) )
	{
		MyLog.Append( "Pixel format could not be set\n" );
		KillWindow();
		return FALSE;
	}
	MyLog.Append( "Pixel format set\n" );

	// Create rendering context
	if( !( hRC = wglCreateContext( hDC ) ) )
	{
		MyLog.Append( "Could not create the rendering context\n" );
		KillWindow();
		return FALSE;
	}
	MyLog.Append( "Rendering context created\n" );

	// Activate rendering context
	if( !wglMakeCurrent( hDC, hRC ) )
	{
		MyLog.Append( "Could not activate the rendering context\n" );
		KillWindow();
		return FALSE;
	}
	MyLog.Append( "Rendering context activated\n" );

	// Show window
	ShowWindow( hWnd, SW_SHOW );
	SetForegroundWindow( hWnd );
	SetFocus( hWnd );

	// Set perspective
	ResizeScene( iWidth, iHeight );

	// Initialize OpenGL
	if( !Initialize() )
	{
		KillWindow();
		return FALSE;
	}
	return TRUE;
}

BOOL WinampKey( int iKey )
{
	#define VK_B 66
	#define VK_C 67
	#define VK_V 86
	#define VK_X 88
	#define VK_Z 90

	switch( iKey )
	{
		// Play/Restart/Unpause
		case VK_X:
		case VK_NUMPAD5:

		// Stop
		case VK_V:

		// Pause/Unpause
		case VK_C:

		// Next track
		case VK_B:
		case VK_NUMPAD6:

		// Previous track
		case VK_Z:
		case VK_NUMPAD4:
		
		// Rewind 5 seconds
		case VK_LEFT:
		case VK_NUMPAD7:

		// Fast-forward 5 seconds
		case VK_RIGHT:
		case VK_NUMPAD9:

		// Jump Ten Songs Back
		case VK_NUMPAD1:

		// Jump Ten Songs Forward
		case VK_NUMPAD3:

		// Turn Volume Up
		case VK_UP:
		case VK_NUMPAD8:

		// Turn Volume Down
		case VK_DOWN:
		case VK_NUMPAD2:
		{
			return TRUE;
			// break;
		}
		default:
		{
			return FALSE;
			// break;
		}
	}
}

LRESULT CALLBACK WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
	switch( message )
	{
		case WM_CREATE:
		{
			if( MySaver.iMode == MODE_FULLSCREEN )
			{
				// Save initial position		
				GetCursorPos( &MousePosFirst );
				MyLog.Append( "Initial mouse position saved\n" );

				//Hide mouse pointer
				ShowCursor( FALSE );
				MyLog.Append( "Mouse pointer hidden\n" );
				return FALSE;
			}
			break;
		}
		case WM_LBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:
		{
			// Reactivate
			if( bPowersaveMonitor )
			{
				bActive = TRUE;
				bPowersaveMonitor = FALSE;
				MyTimer.Continue();
				MyLog.Append( "Animation continued\n" );
			}

			if( MySaver.iMode == MODE_FULLSCREEN )
			{
				PostMessage( hWnd, WM_CLOSE, 0, 0 );
				MyLog.Append( "Mousebutton pressed\n" );
				return FALSE;
			}
			break;
		}
		case WM_MOUSEWHEEL:
		{
			// Reactivate
			if( bPowersaveMonitor )
			{
				bActive = TRUE;
				bPowersaveMonitor = FALSE;
				MyTimer.Continue();
				MyLog.Append( "Animation continued\n" );
			}

			if( MySaver.iMode == MODE_FULLSCREEN )
			{
				PostMessage( hWnd, WM_CLOSE, 0, 0 );
				MyLog.Append( "Mousewheel rotated\n" );
				return FALSE;
			}
			break;
		}
		case WM_MOUSEMOVE:
		{
			// Reactivate
			if( bPowersaveMonitor )
			{
				bActive = TRUE;
				bPowersaveMonitor = FALSE;
				MyTimer.Continue();
				MyLog.Append( "Animation continued\n" );
			}

			if( MySaver.iMode == MODE_FULLSCREEN )
			{
				POINT MousePosCurrent;
				GetCursorPos( &MousePosCurrent );

				if( ( abs( MousePosFirst.x - MousePosCurrent.x ) >= MIN_MOUSE_MOVE ) || ( abs( MousePosFirst.y - MousePosCurrent.y ) >= MIN_MOUSE_MOVE ) )
				{
					PostMessage( hWnd, WM_CLOSE, 0, 0 );
					MyLog.Append( "Mousemove detected\n" );
					return FALSE;
				}
			}
			break;
		}
		case WM_KEYDOWN:
		{
			// Reactivate
			if( bPowersaveMonitor )
			{
				bActive = TRUE;
				bPowersaveMonitor = FALSE;
				MyTimer.Continue();
				MyLog.Append( "Animation continued\n" );
			}

			// Process key
			if( WinampKey( wParam ) )
			{
				// Winamp key
				if( bWinampFound )
				{
					PostMessage( hWinamp, WM_KEYDOWN, wParam, lParam );
				}
			}
			else
			{
				// Internal key

				// Test bit 30
				if( ( lParam & 1073741824 ) == 0 )
				{
					// Key was up before
					switch( wParam )
					{
						// Pause
						case VK_P:
						case VK_PAUSE:
						{
							bActive = !bActive;
							if( bActive )
							{
								MyTimer.Continue();
								MyLog.Append( "Animation continued\n" );
							}
							else
							{
								MyTimer.Pause();
								MyLog.Append( "Animation interrupted\n" );
							}
							break;
						}

						// Transparency
						case VK_T:
						{
							if( bActive )
							{
								// Switch blending mode
								bBlending = !bBlending;
								EnableBlending( bBlending );
								ChooseDepthTestFunction( bBlending );
							}
						}
					}
				}
			}

			return FALSE;
			// break;
		}
		case WM_SYSKEYDOWN:
		{
			// Reactivate
			if( bPowersaveMonitor )
			{
				bActive = TRUE;
				bPowersaveMonitor = FALSE;
				MyTimer.Continue();
				MyLog.Append( "Animation continued\n" );
			}

			return FALSE;
			// break;
		}
		case WM_SYSCOMMAND:
		{
			if( MySaver.iMode == MODE_FULLSCREEN )
			{
				switch( wParam )
				{
  					// Deny screensavers
					case SC_SCREENSAVE:
					{
						MyLog.Append( "Screensaver start prevented\n" );
						return FALSE; // TRUE?
					}

					// Interrupt on powersave
					case SC_MONITORPOWER:
					{
						switch( lParam )
						{
							case 1:
							{
								MyLog.Append( "Monitor switched to powersave\n" );
								break;
							}
							case 2:
							{
								MyLog.Append( "Monitor shut off\n" );
							}
						}

						if( !bPowersaveMonitor )
						{
							bActive = FALSE;
							bPowersaveMonitor = TRUE;
							MyTimer.Pause();
							MyLog.Append( "Animation interrupted\n" );
						}

						break;
					}

					// Deny [ALT]+[F4]
					case SC_CLOSE:
					{
						MyLog.Append( "[ALT]+[F4] prevented\n" );
						return FALSE;
					}
				}
			}
			break;
		}

		case WM_DESTROY:
		{
			// Reveal mouse pointer
			ShowCursor( TRUE );
			MyLog.Append( "Mouse pointer revealed\n" );

			// Time to quit
			PostQuitMessage( 0 );
			return FALSE;
			// break;
		}

		case WM_POWERBROADCAST:
		{
			switch( wParam )
			{
				case PBT_APMQUERYSUSPEND:
				{
					// Suspension is requested
					MyLog.Append( "Suspension request granted\n" );

					// Allow suspension
					return TRUE;
					// break;
				}
				case PBT_APMBATTERYLOW:
				{
					// Battery is low
					PostMessage( hWnd, WM_CLOSE, 0, 0 );
					MyLog.Append( "Low battery detected\n" );
					break;
				}
				case PBT_APMSUSPEND:
				{
					// Suspension is about to be entered
					if( !bPowersaveSuspension )
					{
						bActive = FALSE;
						bPowersaveSuspension = TRUE;
						MyTimer.Pause();
						MyLog.Append( "Suspensed state entered\n" );
						MyLog.Append( "Animation interrupted\n" );
					}
					break;
				}
				// case PBT_APMQUERYSUSPENDFAILED:
				// case PBT_APMRESUMEAUTOMATIC:
				case PBT_APMRESUMESUSPEND:
				{
					// Resume operation
					if( bPowersaveSuspension )
					{
						bActive = TRUE;
						bPowersaveSuspension = FALSE;
						MyTimer.Continue();
						MyLog.Append( "Operation resumed\n" );
						MyLog.Append( "Animation continued\n" );
					}
					break;					
				}
				case PBT_APMPOWERSTATUSCHANGE:
				{
					// Detect power status
					SYSTEM_POWER_STATUS PowerStatus;
					GetSystemPowerStatus( &PowerStatus );
					if( PowerStatus.ACLineStatus != 1 )
					{
						// Offline or unknown
						switch( PowerStatus.BatteryFlag )
						{
							case 2:		// Low
							case 4:		// Critical
							{
								PostMessage( hWnd, WM_CLOSE, 0, 0 );
								MyLog.Append( "Low battery detected\n" );
								// break;
							}
						}
					}
					break;
				}
			}
		}
	}
	return DefWindowProc( hWnd, message, wParam, lParam );
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow )
{
	MSG		msg;				// Message structure
	BOOL	bQuit = FALSE;		// Quit flag


	// Deny more than one instance
	HANDLE hMutex;
	if( ( MySaver.iMode == MODE_FULLSCREEN ) || ( MySaver.iMode == MODE_PREVIEW ) )
	{
		hMutex = CreateMutex( NULL, TRUE, "ONE_APP_ONE_TIME" );
		if( GetLastError() == ERROR_ALREADY_EXISTS )
		{
			// Existing instance found
			MyLog.Append( "Existing instance found\n" );
			return 0;
		}
	}

	// Initialize timer
	MyTimer.fAnimationStep100 = ANIMATION_STEP_DEFAULT;
	MyTimer.fRotationStep100 = ROTATION_STEP_DEFAULT;
	MyTimer.pAnimationStep = &fAnimationStep;
	MyTimer.pRotationStep = &fRotationStep;

	switch( MySaver.iMode )
	{
		case MODE_FULLSCREEN:
		{
			MyLog.Append( "FULLSCREEN mode entered\n" );

			// Enter screensaver mode
			SystemParametersInfo( SPI_SCREENSAVERRUNNING, TRUE, NULL, 0 );
			MyLog.Append( "SCREENSAVER mode entered\n" );

			// WINAMP
			if( NULL != ( hWinamp = FindWindow( "Winamp v1.x", NULL ) ) )
			{
				bWinampFound = TRUE;
				int iWinampVersion = SendMessage( hWinamp, WM_WA_IPC, 0, IPC_GETVERSION );
				iWinampVersionMajor = ( 65535 & iWinampVersion ) >> 12;											// 2^16-1
				iWinampVersionMinor = 10 * ( ( 4095 & iWinampVersion ) >> 8 ) + ( 15 & iWinampVersion );		// 2^12-1  2^4-1
				MyLog.Append( "Running Winamp %i.%i (%i) detected\n", iWinampVersionMajor, iWinampVersionMinor, iWinampVersion );
			}
			else
			{
				MyLog.Append( "No running Winamp detected\n" );
			}

			break;
		}
		case MODE_PREVIEW:
		{
			MyLog.Append( "PREVIEW mode entered\n" );
			break;
		}
		case MODE_CONFIG:
		{
			MyLog.Append( "CONFIG mode entered\n" );
			ShellAbout( MySaver.hParent, APP_TITLE, "            HARTWORK PROJECT\n                 www.hartwork.org", LoadIcon( NULL, IDI_WINLOGO ) );

			return 0;
		}
		case MODE_PASSWORD:
		{
			MyLog.Append( "PASSWORD mode entered\n" );
			MessageBox( MySaver.hParent, "Password protection not implemented yet.\n", "Password protection", MB_ICONINFORMATION | MB_APPLMODAL );
			return 0;
		}
	}

	// Save instance
	hGlobalInstance = hInstance;

	// Create main window
	OpenWindow();

	// Main loop
	MyLog.Append( "Animation activated\n" );
	while( !bQuit )
	{
		if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
		{
			if( msg.message == WM_QUIT )
			{
				bQuit = TRUE;
			}
			else
			{
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}
		else
		{
			if( bActive )
			{
				SaverEngine();
			}
		}
	}
	MyLog.Append( "Animation deactivated\n" );

	// Get framerate
	MyLog.Append( "Average framerate: %.2ffps\n", MyTimer.iFrames / MyTimer.GetElapsed() );
	MyLog.Append( "Speed factor: %.4fx\n", *MyTimer.pAnimationStep / MyTimer.fAnimationStep100 );

	// Kill window
	KillWindow();

	// Exit screensaver mode
	if( MySaver.iMode == MODE_FULLSCREEN )
	{
		SystemParametersInfo( SPI_SCREENSAVERRUNNING, FALSE, NULL, 0 );
		MyLog.Append( "SCREENSAVER mode left\n" );
	}

	// Release mutex
	ReleaseMutex( hMutex );

	return msg.wParam;
}
